@@ -18,6 +18,9 @@ module Database.Persist.TH
     , mpsBackend
     , mpsGeneric
     , mpsPrefixFields
+    , mpsEntityJSON
+    , mpsGenerateLenses
+    , EntityJSON(..)
     , mkPersistSettings
     , sqlSettings
     , sqlOnlySettings
@@ -28,12 +31,17 @@ module Database.Persist.TH
     , share
     , derivePersistField
     , persistFieldFromEntity
+      -- * Internal
+    , packPTH
+    , lensPTH
     ) where
 
 import Prelude hiding ((++), take, concat, splitAt)
+import qualified Prelude as P 
 import Database.Persist
 import Database.Persist.Sql (Migration, SqlPersistT, migrate, SqlBackend, PersistFieldSql)
 import Database.Persist.Quasi
+import Language.Haskell.TH.Lib (varE)
 import Language.Haskell.TH.Quote
 import Language.Haskell.TH.Syntax
 import Data.Char (toLower, toUpper)
@@ -47,6 +55,7 @@ import Data.List (foldl', find)
 import Data.Maybe (isJust)
 import Data.Monoid (mappend, mconcat)
 import qualified Data.Map as M
+import qualified Data.HashMap.Strict as HM
 import Data.Aeson
     ( ToJSON (toJSON), FromJSON (parseJSON), (.=), object
     , Value (Object), (.:), (.:?)
@@ -177,6 +186,29 @@ data MkPersistSettings = MkPersistSettings
     -- True.
     , mpsPrefixFields :: Bool
     -- ^ Prefix field names with the model name. Default: True.
+    , mpsEntityJSON :: Maybe EntityJSON
+    -- ^ Generate @ToJSON@/@FromJSON@ instances for each model types. If it's
+    -- @Nothing@, no instances will be generated. Default:
+    --
+    -- @
+    --  Just EntityJSON
+    --      { entityToJSON = 'keyValueEntityToJSON
+    --      , entityFromJSON = 'keyValueEntityFromJSON
+    --      }
+    -- @
+    , mpsGenerateLenses :: !Bool
+    -- ^ Instead of generating normal field accessors, generator lens-style accessors.
+    --
+    -- Default: False
+    --
+    -- Since 1.3.1
+    }
+
+data EntityJSON = EntityJSON
+    { entityToJSON :: Name
+    -- ^ Name of the @toJSON@ implementation for @Entity a@.
+    , entityFromJSON :: Name
+    -- ^ Name of the @fromJSON@ implementation for @Entity a@.
     }
 
 -- | Create an @MkPersistSettings@ with default values.
@@ -186,6 +218,11 @@ mkPersistSettings t = MkPersistSettings
     { mpsBackend = t
     , mpsGeneric = True -- FIXME switch default to False in the future
     , mpsPrefixFields = True
+    , mpsEntityJSON = Just EntityJSON
+        { entityToJSON = 'keyValueEntityToJSON
+        , entityFromJSON = 'keyValueEntityFromJSON
+        }
+    , mpsGenerateLenses = False
     }
 
 -- | Use the 'SqlPersist' backend.
@@ -198,11 +235,19 @@ sqlSettings = mkPersistSettings $ ConT ''SqlBackend
 sqlOnlySettings :: MkPersistSettings
 sqlOnlySettings = sqlSettings { mpsGeneric = False }
 
-recName :: MkPersistSettings -> Text -> Text -> Text
-recName mps dt f
+recNameNoUnderscore :: MkPersistSettings -> Text -> Text -> Text
+recNameNoUnderscore mps dt f
   | mpsPrefixFields mps = lowerFirst dt ++ upperFirst f
   | otherwise           = lowerFirst f
 
+recName :: MkPersistSettings -> Text -> Text -> Text
+recName mps dt f =
+    addUnderscore $ recNameNoUnderscore mps dt f
+  where
+    addUnderscore
+        | mpsGenerateLenses mps = ("_" ++)
+        | otherwise = id
+
 lowerFirst :: Text -> Text
 lowerFirst t =
     case uncons t of
@@ -483,12 +528,12 @@ mkFromPersistValues mps t@(EntityDef { entitySum = True }) = do
 
 type Lens s t a b = forall f. Functor f => (a -> f b) -> s -> f t
 
-lens :: (s -> a) -> (s -> b -> t) -> Lens s t a b
-lens sa sbt afb s = fmap (sbt s) (afb $ sa s)
+lensPTH :: (s -> a) -> (s -> b -> t) -> Lens s t a b
+lensPTH sa sbt afb s = fmap (sbt s) (afb $ sa s)
 
 mkLensClauses :: MkPersistSettings -> EntityDef a -> Q [Clause]
 mkLensClauses mps t = do
-    lens' <- [|lens|]
+    lens' <- [|lensPTH|]
     getId <- [|entityKey|]
     setId <- [|\(Entity _ value) key -> Entity key value|]
     getVal <- [|entityVal|]
@@ -549,7 +594,8 @@ mkEntity mps t = do
     fpv <- mkFromPersistValues mps t
     utv <- mkUniqueToValues $ entityUniques t
     puk <- mkUniqueKeys t
-
+    fkc <- mapM (mkForeignKeysComposite mps t) $ entityForeigns t
+    
     fields <- mapM (mkField mps t) $ FieldDef
         { fieldHaskell = HaskellName "Id"
         , fieldDB = entityID t
@@ -570,9 +616,11 @@ mkEntity mps t = do
 
     lensClauses <- mkLensClauses mps t
 
-    return $ addSyn
-      [ dataTypeDec mps t
-      , TySynD (mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") [] $
+    lenses <- mkLenses mps t
+
+    return $ addSyn $
+       dataTypeDec mps t : mconcat fkc `mappend`
+      ([ TySynD (mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") [] $
             ConT ''KeyBackend `AppT` mpsBackend mps `AppT` ConT (mkName nameS)
       , InstanceD [] clazz $
         [ uniqueTypeDec mps t
@@ -604,27 +652,105 @@ mkEntity mps t = do
         , FunD 'persistIdField [Clause [] (NormalB $ ConE $ mkName $ unpack $ unHaskellName (entityHaskell t) ++ "Id") []]
         , FunD 'fieldLens lensClauses
         ]
-      ]
+      ] `mappend` lenses)
+
+mkLenses :: MkPersistSettings -> EntityDef a -> Q [Dec]
+mkLenses mps _ | not (mpsGenerateLenses mps) = return []
+mkLenses _ ent | entitySum ent = return []
+mkLenses mps ent = fmap mconcat $ forM (entityFields ent) $ \field -> do
+    let lensName' = recNameNoUnderscore mps (unHaskellName $ entityHaskell ent) (unHaskellName $ fieldHaskell field)
+        lensName = mkName $ unpack lensName'
+        fieldName = mkName $ unpack $ "_" ++ lensName'
+    needleN <- newName "needle"
+    setterN <- newName "setter"
+    fN <- newName "f"
+    aN <- newName "a"
+    yN <- newName "y"
+    let needle = VarE needleN
+        setter = VarE setterN
+        f = VarE fN
+        a = VarE aN
+        y = VarE yN
+        fT = mkName "f"
+        -- FIXME if we want to get really fancy, then: if this field is the
+        -- *only* Id field present, then set backend1 and backend2 to different
+        -- values
+        backend1 = mkName "backend"
+        backend2 = mkName "backend"
+        aT = pairToType mps backend1 (fieldType field, nullable $ fieldAttrs field)
+        bT = pairToType mps backend2 (fieldType field, nullable $ fieldAttrs field)
+        mkST backend = genericDataType mps (unHaskellName $ entityHaskell ent) (VarT backend)
+        sT = mkST backend1
+        tT = mkST backend2
+        t1 `arrow` t2 = ArrowT `AppT` t1 `AppT` t2
+        vars = PlainTV fT
+             : (if mpsGeneric mps then [PlainTV backend1{-, PlainTV backend2-}] else [])
+    return
+        [ SigD lensName $ ForallT vars [ClassP ''Functor [VarT fT]] $
+            (aT `arrow` (VarT fT `AppT` bT)) `arrow`
+            (sT `arrow` (VarT fT `AppT` tT))
+        , FunD lensName $ return $ Clause
+            [VarP fN, VarP aN]
+            (NormalB $ VarE 'fmap
+                `AppE` setter
+                `AppE` (f `AppE` needle))
+            [ FunD needleN [Clause [] (NormalB $ VarE fieldName `AppE` a) []]
+            , FunD setterN $ return $ Clause
+                [VarP yN]
+                (NormalB $ RecUpdE a
+                    [ (fieldName, y)
+                    ])
+                []
+            ]
+        ]
+
+mkForeignKeysComposite :: MkPersistSettings -> EntityDef a -> ForeignDef -> Q [Dec]
+mkForeignKeysComposite mps t fdef = do
+   let fieldName f = mkName $ unpack $ recName mps (unHaskellName $ entityHaskell t) (unHaskellName f)
+   let fname=fieldName $ foreignConstraintNameHaskell fdef
+   let reftablename=mkName $ unpack $ unHaskellName $ foreignRefTableHaskell fdef 
+   let tablename=mkName $ unpack $ unHaskellName $ entityHaskell t
+   entName <- newName "entname"
+   
+   let flds = map (\(a,_,_,_) -> VarE (fieldName a)) $ foreignFields fdef
+   let xs = ListE $ map (\a -> AppE (VarE 'toPersistValue) ((AppE a (VarE entName)))) flds
+   let fn = FunD fname [Clause [VarP entName] (NormalB (AppE (ConE 'Key) (AppE (ConE 'PersistList) xs))) []]
+   
+   let t2 = ConT ''KeyBackend `AppT` ConT ''SqlBackend `AppT` ConT reftablename
+   let sig = SigD fname $ (ArrowT `AppT` (ConT tablename)) `AppT` t2
+   return [sig, fn]
+
 
 -- | produce code similar to the following:
 --
 -- instance PersistEntity e => PersistField e where
 --    toPersistValue = PersistMap $ zip columNames (map toPersistValue . toPersistFields)
---    fromPersistValue (PersistMap o) = fromPersistValues $ map (\(_,v) ->
---        casefromPersistValue v of
---            Left e -> error e
---            Right r -> r) o
+--    fromPersistValue (PersistMap o) = 
+--        let columns = HM.fromList x
+--        in fromPersistValues $ map (\name ->
+--          case HM.lookup name o of
+--            Just v ->
+--              case fromPersistValue v of
+--                Left e -> error e
+--                Right r -> r
+--            Nothing -> error $ "Missing field: " `mappend` unpack name) columnNames 
 --    fromPersistValue x = Left $ "Expected PersistMap, received: " ++ show x
 --    sqlType _ = SqlString
 persistFieldFromEntity :: MkPersistSettings -> EntityDef a -> Q [Dec]
 persistFieldFromEntity mps e = do
     ss <- [|SqlString|]
-    let columnNames = map (unpack . unHaskellName . fieldHaskell) (entityFields e)
     obj <- [|\ent -> PersistMap $ zip (map pack columnNames) (map toPersistValue $ toPersistFields ent)|]
-    fpv <- [|\x -> fromPersistValues $ map (\(_,v) -> case fromPersistValue v of
-                                                      Left e' -> error $ unpack e'
-                                                      Right r -> r) x|]
-    let typ = genericDataType mps (pack entityName) $ VarT $ mkName "backend"
+    fpv <- [|\x -> let columns = HM.fromList x
+                    in fromPersistValues $ map
+                         (\(nulled, name) ->
+                            case HM.lookup name columns of
+                                Just v -> case fromPersistValue v of
+                                    Left e' -> error $ unpack e'
+                                    Right r -> r
+                                Nothing -> if nulled then PersistNull
+                                    else error $ "Missing field: " `mappend` unpack name)
+                         (zip maybeColumns $ map pack columnNames)
+          |]
 
     compose <- [|(<=<)|]
     getPersistMap' <- [|getPersistMap|]
@@ -640,7 +766,11 @@ persistFieldFromEntity mps e = do
             ]
         ]
     where
+      typ = genericDataType mps (pack entityName) $ VarT $ mkName "backend"
       entityName = (unpack $ unHaskellName $ entityHaskell e)
+      entFields = entityFields e
+      columnNames  = map (unpack . unHaskellName . fieldHaskell) entFields
+      maybeColumns = map ((== Nullable ByMaybeAttr) . nullable . fieldAttrs) entFields
 
 -- | Apply the given list of functions to the same @EntityDef@s.
 --
@@ -834,7 +964,7 @@ mkMigrate fun allDefs = do
         return $ NoBindS $ m `AppE` defsExp `AppE` u
 
 instance Lift' a => Lift (EntityDef a) where
-    lift (EntityDef a b c d e f g h i) =
+    lift (EntityDef a b c d e f g h i j k) =
         [|EntityDef
             $(lift a)
             $(lift b)
@@ -842,14 +972,20 @@ instance Lift' a => Lift (EntityDef a) where
             $(liftTs d)
             $(lift e)
             $(lift f)
-            $(liftTs g)
-            $(liftMap h)
-            $(lift i)
+            $(lift g)
+            $(lift h)
+            $(liftTs i)
+            $(liftMap j)
+            $(lift k)
             |]
 instance Lift' a => Lift (FieldDef a) where
     lift (FieldDef a b c d e f g) = [|FieldDef a b c $(lift' d) $(liftTs e) f $(lift' g)|]
 instance Lift UniqueDef where
     lift (UniqueDef a b c d) = [|UniqueDef $(lift a) $(lift b) $(lift c) $(liftTs d)|]
+instance Lift PrimaryDef where
+    lift (PrimaryDef a b) = [|PrimaryDef $(lift a) $(liftTs b)|]
+instance Lift ForeignDef where
+    lift (ForeignDef a b c d e f) = [|ForeignDef $(lift a) $(lift b) $(lift c) $(lift d) $(lift e) $(liftTs f)|]
 
 -- | A hack to avoid orphans.
 class Lift' a where
@@ -866,14 +1002,14 @@ instance Lift' () where
 instance Lift' SqlTypeExp where
     lift' = lift
 
-pack' :: String -> Text
-pack' = pack
+packPTH :: String -> Text
+packPTH = pack
 #if !MIN_VERSION_text(0, 11, 2)
-{-# NOINLINE pack' #-}
+{-# NOINLINE packPTH #-}
 #endif
 
 liftT :: Text -> Q Exp
-liftT t = [|pack' $(lift (unpack t))|]
+liftT t = [|packPTH $(lift (unpack t))|]
 
 liftTs :: [Text] -> Q Exp
 liftTs = fmap ListE . mapM liftT
@@ -923,8 +1059,8 @@ instance Lift SqlType where
     lift (SqlNumeric x y) =
         [|SqlNumeric (fromInteger x') (fromInteger y')|]
       where
-        x' = fromIntegral x
-        y' = fromIntegral y
+        x' = fromIntegral x :: Integer
+        y' = fromIntegral y :: Integer
     lift SqlBool = [|SqlBool|]
     lift SqlDay = [|SqlDay|]
     lift SqlTime = [|SqlTime|]
@@ -1044,4 +1180,13 @@ mkJSON mps def = do
             (Just $ VarE obj)
             (if nullable (fieldAttrs f) == Nullable ByMaybeAttr then dotColonQE else dotColonE)
             (Just $ AppE packE $ LitE $ StringL $ unpack $ unHaskellName $ fieldHaskell f)
-    return [toJSONI, fromJSONI]
+    case mpsEntityJSON mps of
+        Nothing -> return [toJSONI, fromJSONI]
+        Just entityJSON -> do
+            entityJSONIs <- [d|
+                instance ToJSON (Entity $(pure typ)) where
+                    toJSON = $(varE (entityToJSON entityJSON))
+                instance FromJSON (Entity $(pure typ)) where
+                    parseJSON = $(varE (entityFromJSON entityJSON))
+                |]
+            return $ toJSONI : fromJSONI : entityJSONIs
